:- module(sq, [squares_clpfd/5]).

:- use_module(library(clpfd)).
:- use_module(library(lists)).

squares_clpfd(Flags-Over-Lab, Sizes, Limit, Xs, Ys) :-
        generate_coordinates(Xs, Ys, Sizes, Limit),
	state_asymmetry(Xs, Ys, Sizes, Limit),
	state_no_overlap(Xs, Ys, Sizes, Over),
	(   member(cum, Flags)
	->  cumulative(Xs, Sizes, Sizes, Limit),
	    cumulative(Ys, Sizes, Sizes, Limit)
	;   true
	),
	(   member(cap, Flags)
	->  state_capacity(Xs, Sizes, Limit),
	    state_capacity(Ys, Sizes, Limit)
	;   true
	),
	(   Lab = dual
	->  dual_labeling(Xs, 1, Limit), 
	    dual_labeling(Ys, 1, Limit)
	;   labeling(Lab, Xs),
	    labeling(Lab, Ys)
	).

generate_coordinates([], [], [], _).
generate_coordinates([X|Xs], [Y|Ys], [S|Ss], Limit) :-
        Sd is Limit-S+1,
        domain([X,Y], 1, Sd),
        generate_coordinates(Xs, Ys, Ss, Limit).

% first square has center in SW quarter, under the positive diagonal
state_asymmetry([X|_], [Y|_], [D|_], Limit) :-
	UB is (Limit-D+2)>>1,
	X in 1..UB,
	Y #=< X.

state_no_overlap([], [], [], _).
state_no_overlap([X|Xs], [Y|Ys], [S|Ss], Type) :-
	state_no_overlap(X, Y, S, Xs, Ys, Ss, Type),
	state_no_overlap(Xs, Ys, Ss, Type).

state_no_overlap(_, _, _, [], [], [], _).
state_no_overlap(X, Y, S, [X1|Xs], [Y1|Ys], [S1|Ss], Type) :-
	no_overlap(Type, X, Y, S, X1, Y1, S1),
	state_no_overlap(X, Y, S, Xs, Ys, Ss, Type).

no_overlap(spec, X1, _Y1, S1, X2, _Y2, _S2) :-
    	X1+S1 #=< X2.
no_overlap(spec, X1, _Y1, _S1, X2, _Y2, S2) :-
    	X2+S2 #=< X1.
no_overlap(spec, _X1, Y1, S1, _X2, Y2, _S2) :-
    	Y1+S1 #=< Y2.
no_overlap(spec, _X1, Y1, _S1, _X2, Y2, S2) :-
    	Y2+S2 #=< Y1.

no_overlap(card1, X1, Y1, S1, X2, Y2, S2) :-
	X1+S1 #=< X2 #<=> B1,
	X2+S2 #=< X1 #<=> B2,
	Y1+S1 #=< Y2 #<=> B3,
	Y2+S2 #=< Y1 #<=> B4,
	B1+B2+B3+B4 #>= 1.

no_overlap(card2, X1, Y1, S1, X2, Y2, S2) :-
	call( abs(2*(X1-X2)+(S1-S2)) #>= S1+S2 #\/
	    abs(2*(Y1-Y2)+(S1-S2)) #>= S1+S2 ).

no_overlap(wcd, X1, Y1, S1, X2, Y2, S2) :-
	no_overlap_ix(X1, Y1, S1, X2, Y2, S2).

/*
no_overlap_ix(X1, Y1, S1, X2, Y2, S2) +:
  X1 in ((((min(Y1)+S1)..max(Y2))\/((min(Y2)+S2)..max(Y1))) ? (inf..sup))
        \/ ((inf..max(X2)-S1) \/ (min(X2)+S2..sup)),
  X2 in ((((min(Y1)+S1)..max(Y2))\/((min(Y2)+S2)..max(Y1))) ? (inf..sup))
        \/ ((inf..max(X1)-S2) \/ (min(X1)+S1..sup)),
  Y1 in ((((min(X1)+S1)..max(X2))\/((min(X2)+S2)..max(X1))) ? (inf..sup))
        \/ ((inf..max(Y2)-S1) \/ (min(Y2)+S2..sup)),
  Y2 in ((((min(X1)+S1)..max(X2))\/((min(X2)+S2)..max(X1))) ? (inf..sup))
        \/ ((inf..max(Y1)-S2) \/ (min(Y1)+S1..sup)).
*/

no_overlap_ix(X1, Y1, S1, X2, Y2, S2) +:
  X1 in ((((min(Y1)+S1)..max(Y2))\/((min(Y2)+S2)..max(Y1))) ? (inf..sup))
	\/ \(max(X2)-(S1-1) .. min(X2)+(S2-1)),
  X2 in ((((min(Y1)+S1)..max(Y2))\/((min(Y2)+S2)..max(Y1))) ? (inf..sup))
	\/ \(max(X1)-(S2-1) .. min(X1)+(S1-1)),
  Y1 in ((((min(X1)+S1)..max(X2))\/((min(X2)+S2)..max(X1))) ? (inf..sup))
	\/ \(max(Y2)-(S1-1) .. min(Y2)+(S2-1)),
  Y2 in ((((min(X1)+S1)..max(X2))\/((min(X2)+S2)..max(X1))) ? (inf..sup))
	\/ \(max(Y1)-(S2-1) .. min(Y1)+(S1-1)).




dual_labeling([], _, _) :- !.
dual_labeling(L, I, Limit) :-
        dual_labeling(L, L1, I, Limit, J),
        dual_labeling(L1, J, Limit).

dual_labeling([], [], _, J, J).
dual_labeling([X|L1], L2, I, J0, J) :-
        (   integer(X) -> dual_labeling(L1, L2, I, J0, J)
        ;   X = I, dual_labeling(L1, L2, I, J0, J)
        ;   X #> I,
            fd_min(X, J1), J2 is min(J0,J1), L2 = [X|L3],
            dual_labeling(L1, L3, I, J2, J)
        ).

state_capacity(Cs, Sizes, Limit) :-
        state_capacity(1, Limit, Cs, Sizes).

state_capacity(Pos, Limit, _, _) :- 
        Pos>Limit, !.
state_capacity(Pos, Limit, Cs, Sizes) :-
    	accumulate(Cs, Sizes, Pos, Bs),
	scalar_product(Sizes, Bs, #=, Limit),
	Pos1 is Pos+1,
        state_capacity(Pos1, Limit, Cs, Sizes).

accumulate([], [], _, []).
accumulate([C|Cs], [S|Ss], Pos, [B|Bs]) :-
	Crutch is Pos-S+1,
	C in Crutch .. Pos #<=> B,
	accumulate(Cs, Ss, Pos, Bs).


no_overlap(X1, Y1, S1, X2, Y2, S2) +:
  X1 in ((((min(Y1)+S1)..max(Y2))\/((min(Y2)+S2)..max(Y1))) ? (inf..sup))
        \/ ((inf..max(X2)-S1) \/ (min(X2)+S2..sup)),
  X2 in ((((min(Y1)+S1)..max(Y2))\/((min(Y2)+S2)..max(Y1))) ? (inf..sup))
        \/ ((inf..max(X1)-S2) \/ (min(X1)+S1..sup)),
  Y1 in ((((min(X1)+S1)..max(X2))\/((min(X2)+S2)..max(X1))) ? (inf..sup))
        \/ ((inf..max(Y2)-S1) \/ (min(Y2)+S2..sup)),
  Y2 in ((((min(X1)+S1)..max(X2))\/((min(X2)+S2)..max(X1))) ? (inf..sup))
        \/ ((inf..max(Y1)-S2) \/ (min(Y1)+S1..sup)).
