:- ensure_loaded(sq_clpfd).
:- ensure_loaded(sq_prolog).

:- use_module(library(timeout)).
:- use_module(library(clpfd)).

time(Id, Goal) :-
	init_backtracks(Id),
	statistics(runtime,_),
	Goal, !,
	statistics(runtime, [_, T]),
	get_backtracks(Id, BType, B),
	show_result(Id, T, BType, B).
time(Id, _Goal):- 
	statistics(runtime, [_, T]), 
	get_backtracks(Id, BType, B),
	show_result(Id/'(failed)', T, BType, B).


squares(Flags-Over-Lab, Sizes, Limit) :-
	squares_clpfd(Flags-Over-Lab, Sizes, Limit, _Xs, _Ys).
squares(prolog, Sizes, Limit) :-
	squares_prolog(Sizes, Limit, _SXYs).

init_backtracks(_:prolog) :-
	!, bb_put(count, 0).
init_backtracks(_) :-
	fd_statistics(backtracks, _).

get_backtracks(_:prolog, pl, B) :-
	!, bb_get(count, B).
get_backtracks(_, fd, B) :-
	fd_statistics(backtracks,B).

show_result(Id, T, BType, B) :-
	format(user_error, 
         'Goal: ~|~w,~25+ time: ~|~t~3d~7+ sec, ~a btracks: ~|~t~d~5+\n', [Id,T,BType,B]),
	flush_output(user_error).
	
runt(Goal, Sec) :-
	MSec is Sec*1000, time_out(Goal, MSec, Result),
	(
	    Result = time_out -> 
	    format(user_error, 'Timed out after ~w sec.~n~n', [Sec]),
	    flush_output(user_error)	
	;
	    true
	).

time(Id, Goal, Sec) :-
	runt(time(Id, Goal), Sec).

sq(Id, Limit) :-
	sqd(Limit, Sizes),
	runt(time(Limit:Id, squares(Id, Sizes, Limit)), 200).

sqall :-
	test_case(Max, Id),
	sqd(Limit, _),
	Limit =< Max,	
	sq(Id, Limit), fail.
sqall.

sqd(10,  [6,4,4,4,2,2,2,2]).
sqd(20,  [9,8,8,7,5,4,4,4,4,4,3,3,3,2,2,1,1]).
sqd(112, [50,42,37,35,33,29,27,25,24,19,18,17,16,15,11,9,8,7,6,4,2]).
sqd(175, [81,64,56,55,51,43,39,38,35,33,31,30,29,20,18,16,14,9,8,5,4,3,2,1]).

test_case(112, prolog).

test_case(10, []-spec-[]).
test_case(10, [cap]-spec-[]).
test_case(20, [cap]-card1-[]).
test_case(20, [cap]-card2-[]).
test_case(20, [cap]-wcd-[]).
test_case(10, []-spec-dual).

test_case(10,  [cap]-spec-dual).
test_case(999, [cap]-card1-dual).
test_case(999, [cap]-card2-dual).
test_case(999, [cap]-wcd-dual).

test_case(10,  [cum]-spec-dual).
test_case(999, [cum]-card1-dual).
test_case(999, [cum]-card2-dual).
test_case(999, [cum]-wcd-dual).

test_case(10,  [cap,cum]-spec-dual).
test_case(999, [cap,cum]-card1-dual).
test_case(999, [cap,cum]-card2-dual).
test_case(999, [cap,cum]-wcd-dual).

:- sqall.
