:- module sq.
:- interface.
:- import_module io.

:- pred main(io__state::di, io__state::uo) is cc_multi.

:- implementation.

:- import_module list, int, std_util, string.

main --> 
    io__command_line_arguments(Args),
    (
	{ Args = [Size0|Ss0] },	{ string__to_int(Size0, Size) },
	{ list__map(string__to_int, Ss0, Ss) } ->
	(
	    { squares(Ss, Size, SXYs) } -> write_sols(SXYs)
	;
	    io__write_string("No solution\n")
	)
    ;                 
        io__write_string("Usage: sq <size> <sq1> <sq2> ... \n")
    ).

:- pred write_sols(list(placement)::in, io__state::di, 
           io__state::uo) is det.
write_sols([]) -->
    io__write_string("\n").
write_sols([s(S,X,Y)|Ss]) -->
    io__write_string("square size "), io__write_int(S), 
    io__write_string(" placed at ("), io__write_int(X), 
    io__write_string(","), io__write_int(Y), 
    io__write_string(")\n"), write_sols(Ss).

:- type placement ---> s(size, xcoord, ycoord).
:- type size == int.
:- type xcoord == int.
:- type ycoord == int.
:- type segment ---> xcoord/ycoord.
:- type line == list(segment).
:- type point == pair(xcoord, ycoord).

:- pred squares(list(size), size, list(placement)).
:- mode squares(in, in, out) is nondet.
squares(Ss, Size, SXYs) :-
        MSize is -Size,
        squares(Ss, [MSize/Size,Size/0], Size, [], SXYs).

:- pred squares(list(size)::in, line::in, size::in, 
	list(placement)::in, list(placement)::out) is nondet.
squares(Ss0, L0, Size, SXYs0, SXYs) :-
	(   Ss0=[] -> SXYs=SXYs0
	;   L0=[V0/_|_], Y0 is Size+1+V0,
	    hole(L0, 1-Y0, XY, L1, L2),
	    fill_hole(L2, L3, XY, Size, Ss0, Ss, SXYs0, SXYs1),
	    list__append(L1, L3, L),
	    squares(Ss, L, Size, SXYs1, SXYs)
	).

:- pred hole(line::in, point::in, point::out, 
	     line::out, line::out) is semidet.
hole(LL0, XY0, XY, L1, L2) :-
	LL0=[V1/H1|LL1], LL1=[V2/H2|L],
	(   H1=0 -> V is V1+V2,
	    hole([V/H2|L], XY0, XY, L1, L2)
	;   V2=0 -> H is H1+H2,
	    hole([V1/H|L], XY0, XY, L1, L2)
	;   H1>0, V1<0, V2>0 -> L1=[], L2=LL0, XY=XY0
	;   XY0=X0-Y0, X1 is X0+H1, Y1 is Y0+V2,
	    L1=[V1/H1|L11],
	    hole(LL1, X1-Y1, XY, L11, L2)
	).

:- pred fill_hole(line::in, line::out, point::in, size::in, 
        list(size)::in, list(size)::out, 
	list(placement)::in, list(placement)::out) is nondet.
fill_hole([V0/0,V/H|L], [V1/H|L], _XY, _Size, 
                       Ss0, Ss0, SXYs0, SXYs0) :- 
        !, V1 is  V0+V.
fill_hole([V/H|L0], [V1/S|L1], X-Y, Size,  Ss0, Ss,
                       SXYs0, [s(S,X,Y)|SXYs]) :-
	H > 0, Size1 is Size+1-Y,
        int__min(H, Size1, MaxS),
	select_lseq(S, MaxS, Ss0, Ss1),
        V1 is V+S, MS is -S, H1 is H-S, X1 is X+S, 
        fill_hole([MS/H1|L0], L1, X1-Y, Size, Ss1, Ss, SXYs0, SXYs).

:- pred select_lseq(int::out, int::in, 
	            list(int)::in, list(int)::out) is nondet.
select_lseq(S, Mx, [S0|Ss0], Ss) :-
	(
	    S0 > Mx -> Ss = [S0|Ss1],
	    select_lseq(S, Mx, Ss0, Ss1)
	;
	    S0 = S, Ss = Ss0
	;
	    Ss = [S0|Ss1], Mx1 is S0-1,
	    select_lseq(S, Mx1, Ss0, Ss1)
	).

